/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   grouping of DataLayers to the "undo-stack" Data
 */

#include "Data.hh"

#include <algorithm>
#include <functional>

using namespace std;
using namespace GSML;

/****************************
 * Data::ProgressInfo::Info *
 ****************************/

Data::ProgressInfo::Info::Info (ProgressInfo* i_parent)
    : m_text (), m_progress (0)
{
  m_parent = i_parent;
}

Data::ProgressInfo::Info::Info (ProgressInfo* i_parent, const Data::ProgressInfo::Info& i_info)
  : m_text (i_info.m_text), m_progress (i_info.m_progress)
{
  m_parent = i_parent;
}

void Data::ProgressInfo::Info::SetText (const string& n_text)
{
  m_text = n_text;
  m_parent->SignalInfoUpdated.emit (this);
}

void Data::ProgressInfo::Info::SetProgress (double n_progress)
{
  m_progress = n_progress;
  m_parent->SignalInfoUpdated.emit (this);
}
	  
const string& Data::ProgressInfo::Info::Text () const
{
  return m_text;
}

const double Data::ProgressInfo::Info::Progress () const
{
  return m_progress;
}
	  
Data::ProgressInfo::Info& Data::ProgressInfo::Info::operator= (const Data::ProgressInfo::Info& src)
{
  m_text = src.m_text;
  m_progress = src.m_progress;
  m_parent = src.m_parent;
}

/**********************
 * Data::ProgressInfo *
 *********************/

Data::ProgressInfo::ProgressInfo ()
{
}
	
Data::ProgressInfo::~ProgressInfo ()
{
  if (m_infos.size () != 0)
    cout << "~ProgressInfo: WARNING infos deque not empty?!?" << endl;
}

// for very basic access
const string& Data::ProgressInfo::Text ()
{
  return m_infos.back () -> Text ();
}
	
double Data::ProgressInfo::Progress ()
{
  return m_infos.back () -> Progress ();
}
	
Data::ProgressInfo::Info* Data::ProgressInfo::PushInfo (const string& text)
{
  Info* i = new Info (this);
  i->m_text = text;
	  
  m_infos.push_back (i);
	  
  SignalInfoPushed.emit (i);
  return i;
}
	
void Data::ProgressInfo::PopInfo ()
{
  if (m_infos.size () > 0)
    {
      Info* i = m_infos.back ();
      
      m_infos.pop_back ();
      SignalInfoPopped.emit ();
      delete i;
    }
  else
    {
      ExceptEmpty e;
      e.info = "can't pop info - empty";
      throw e;
    }
}

void Data::ProgressInfo::Clear ()
{
  m_infos.clear ();
}

int Data::ProgressInfo::Count () const
{
  return m_infos.size ();
}

Data::ProgressInfo::const_iterator Data::ProgressInfo::begin ()
{
  return m_infos.begin ();
}
	
Data::ProgressInfo::const_iterator Data::ProgressInfo::end ()
{
  return m_infos.end ();
}

void  Data::ProgressInfo::Debug ()
{
  cout << "Data::ProgressInfo::Debug:" << endl;
  
  cout << "  info objects: " << m_infos.size () << endl;
  
  for (iterator i = m_infos.begin (); i != m_infos.end (); ++ i)
    {
      cout << "  " << i - m_infos.begin () << ": " << (*i)->m_text << endl;
    }
}

/********
 * Data *
 ********/

Data::Data (DataFileIndex* i_fileindex,  unsigned int i_max_layers,
	    PCMDatatype* i_sampletype, sample_rate i_samplerate)
{
  
  m_fileindex = i_fileindex;
  max_layers = i_max_layers;
  m_temp_sampletype = i_sampletype;
  m_temp_samplerate = i_samplerate;
  
  // create initial data_layer
  DataLayer* t_layer = new DataLayer ("Ctor Layer", m_fileindex,
				      m_temp_sampletype, m_temp_samplerate);

  m_data.push_back (t_layer);

  // init some stuff
  current_layer = m_data.begin ();
  prepared_layer = 0;
  m_modified = true; // cause a new data instance ...
}

Data::~Data ()
{
  if (prepared_layer != 0)
    {
      cout << "BUG: ~Data: Layer currently prepared ?!?" << endl;
      delete prepared_layer;
    }
  
  // destroy all tracks ...
  del_helper<DataLayer*> p;
  for_each (m_data.begin (), m_data.end (), p);
}

track_index Data::Tracks () const
{
  return (*current_layer) -> Tracks ();
}

sample_index Data::Length () const
{
  return (*current_layer) -> Length ();
}

DataLayer* Data::CurrentLayerPtr ()
{
  return *current_layer;
}

int Data::Layers ()
{
  return m_data.size ();
}

Data::iterator Data::CurrentLayer ()
{
  return current_layer;
}

Data::iterator Data::BeginLayer ()
{
  return m_data.begin ();
}

Data::iterator Data::EndLayer ()
{
  return m_data.end ();
}
      
void Data::SetCurrentLayer (iterator n_layer)
{
  if (prepared_layer != 0) // already prepared ??
    {
      ExceptLogicalError e;
      e.info = "A new layer is already prepared ... ";
      throw e;
    }
  
  if (n_layer == m_data.end () ) // vaild ??
    {
      ExceptIteratorInvaild e;
      e.info = "The iterator points to a not existing layer";
      throw e;
    }
  
  (*current_layer)->Fixate ();
  current_layer = n_layer;
  (*current_layer)->UnFixate ();
  
  SignalCurrentLayerChanged.emit (*current_layer);
}
 
// creates and return a new data_layar which will be connected to
// the wave_plots for the dynamic graphic preview!!
DataLayer* Data::PrepareNewLayer ()
{
#ifdef DEBUG_DATA
  cout << "DATA::PrepareNewLayer:" << endl;
#endif
  
  if (prepared_layer != 0) // already prepared ??
    {
      ExceptLogicalError e;
      e.info = "A new layer is already prepared ... ";
      throw e;
    }
  
  if (progress_info.Count () != 0)
    {
      cout << "BUG: an effect has left an Info object!!!" << endl;
      progress_info.Debug ();
      progress_info.Clear ();
    }

  prepared_layer = new DataLayer (**current_layer);
  prepared_layer->Comment("not set, yet");
  
  SignalNewLayerPrepared.emit (prepared_layer);
  
  return prepared_layer;
}

// fixate the "in work" data layer
void Data::FixateNewLayer ()
{
#ifdef DEBUG_GSML_DATA
  cout << "DATA::FixateNewLayer:" << endl;
#endif
  if (prepared_layer == 0)
    {
      ExceptLogicalError e;
      e.info = "No new layer prepared ... ";
      throw e;
    }
  
  (*current_layer)->Fixate ();
  
  // remove all undo layer between current and end
  if (current_layer != m_data.end () )
    ++ current_layer;
  m_data.erase (current_layer, m_data.end () );
  
  // add layer
  m_data.push_back (prepared_layer);
  
  prepared_layer = 0;
  current_layer = -- m_data.end ();
  m_modified = true;
  
  if (progress_info.Count () != 0)
    {
      cout << "BUG: a function has left an Info object!!! (I fix it ...)" << endl;
      progress_info.Clear ();
    }
  
  // check and remove more than max layer
  delete_if_too_many_layers ();
  
  // emit Signals
  SignalNewLayerFixated.emit ();
  SignalCurrentLayerChanged.emit (*current_layer);
}

// destroy the "in work" data_layer if the effect fails in some way ...
void Data::CancelNewLayer ()
{
#ifdef DEBUG_GSML_DATA
  cout << "DATA::CancelNewLayer:" << endl;
#endif
  if (prepared_layer == 0)
    {
      Data::Except e;
      e.info = "No new layer prepared ... ";
      throw e;
    }
  
  delete prepared_layer;
  prepared_layer = NULL;
  
  if (progress_info.Count () != 0)
    {
      cout << "BUG: an effect has left an Info object!!!" << endl;
      progress_info.Clear ();
    }
  
  // emit Signal
  SignalNewLayerCanceled.emit ();
}

// configure with new options

DataFileIndex* Data::GetDataFileIndex ()
{
  return m_fileindex;
}

PCMDatatype* Data::TempSampletype ()
{
  return m_temp_sampletype;
}

sample_rate Data::TempSamplerate ()
{
  return m_temp_samplerate;
}

void Data::SetTempSampletype (PCMDatatype* n_type)
{
  m_temp_sampletype = n_type;
}

void Data::SetTempSamplerate (sample_rate n_rate)
{
  m_temp_samplerate = n_rate;
}

void Data::SetMaxLayers (unsigned int n_max_layers)
{
  max_layers = n_max_layers;
  delete_if_too_many_layers ();
}

unsigned int Data::MaxLayers ()
{
  return max_layers;
}

bool Data::IsPrepared ()
{
  return prepared_layer != 0;
}

bool Data::IsModified ()
{
  return m_modified;
}

void Data::SetModified (bool n_state)
{
  m_modified = n_state;
}

void Data::Debug ()
{
  cout << "Data::Debug:" << endl;
  
  cout << "  Layer: " << m_data.size () << endl
       << "  Current_DataLayer: " << current_layer - m_data.begin ()<< endl
       << "  Prepared_Layer: " << prepared_layer << endl
       << "  Modified: " << m_modified << endl;
  
  cout << "  config:" << endl
       << "    max_layers: " << max_layers << endl
       << "    file_index: " << m_fileindex << endl
       << "    temp_sampletype: " << m_temp_sampletype << endl
       << "    temp_samplerate: " << m_temp_samplerate << endl;
  
  for (iterator i = m_data.begin (); i != m_data.end (); ++ i)
    {
      cout << "  Layer: " << i - m_data.begin () << " -> Debug:" << endl;
      (*i)->Debug ();
    }
    
  cout << "End ..." << endl;
}

// better than it has be ...
void Data::delete_if_too_many_layers ()
{
  cout << "Data::delete_if_too_many_layers" << endl;
  if ( m_data.size () > max_layers)
    {
      int current_delta;
      int delete_count;
      
#ifdef DEBUG_GSML_DATA
      cout << "more than max_layer (max: " << may_layers
	   << ", now: " << m_data.size () << ")"<< endl;
#endif
      current_delta = current_layer - m_data.begin ();
      delete_count = m_data.size () - max_layers;
      
      if (delete_count > current_delta)
	{
	  cout << "delete_if_too_many_layers: can't delete max ..." << endl;
	  delete_count = current_delta;
	}
      
      m_data.erase (m_data.begin (), m_data.begin () + delete_count);
      
      // adjust the current data layer ...
      current_layer = m_data.begin () + current_delta - delete_count;
    }
}
