/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2002 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Thanks to the Imlib2 authors and to Foley, van Dam, Feiner and
 * Hughe who wrote: "Computer Graphics: Principles and Practices"
 * Without them this new graphic toolkit would have been much harder!
 */

/* Short Description:
 *   An OO, anti-aliased and alpha-blending graphics primitives implementation.
 */

#include <gtk--/style.h>

#include <algorithm> // STL algorithms
#include <functional>

#include "plugins-gtk/Pixmap.hh"

using namespace std;
using namespace GSML;
using namespace GSMP;

/**************
 * GsmpPixmap *
 **************/

GsmpPixmap::GsmpPixmap ()
{
  m_data = 0;
}

GsmpPixmap::GsmpPixmap (coord width, coord height)
{
  m_area.w = width;
  m_area.h = height;
  
  m_data = new GsmpColor::Color [m_area.w * m_area.h];
}

GsmpPixmap::GsmpPixmap (const GsmpPixmap& source)
{
  m_area = source.m_area;
  
  m_data = new GsmpColor::Color [m_area.w * m_area.h];
  
  memcpy (m_data, source.m_data, m_area.w * m_area.h * sizeof (GsmpColor::Color) );
}


GsmpPixmap::~GsmpPixmap ()
{
  if (m_data != 0)
    delete [] m_data;
}

GsmpPixmap::point GsmpPixmap::Size () const
{
  point p;
  p.x = m_area.w;
  p.y = m_area.h;
  
  return p;
}

const GsmpPixmap::area& GsmpPixmap::Area () const
{
  return m_area;
}

void GsmpPixmap::Resize (const GsmpPixmap::point& n_size)
{
  // TODO: bitblt existing image data ...
  
  if (m_data != 0)
    delete [] m_data;
  
  m_area.w = n_size.x;
  m_area.h = n_size.y;
  
  m_data = new GsmpColor::Color [m_area.w * m_area.h];
  cout << "GsmpPixmap::Resize: " << m_area << endl;
}

void GsmpPixmap::Clear (GsmpColor::Color c)
{
  for (coord i = 0; i < m_area.w * m_area.h; ++ i)
    m_data [i] = c;
}

void GsmpPixmap::BitBlt (GsmpPixmap::area source, const GsmpPixmap::point& dest_point)
{
  // clip
  source.Clip (m_area);
  area dest (dest_point.x, dest_point.y, source.w, source.h);
  dest.Clip (m_area); // clip destination boundaries
  
  cout << "GsmpPixmap::BitBlt: source: " << source << ", dest: " << dest << endl;
  
  // calculte bitblt values
  coord x_start, y_start, per_pixel, per_src_line, per_dest_line;
  
  if (source.x == dest.x && source.y == dest.y)
    return;
  
  if (source.x <= dest.x && source.y <= dest.y) {
    x_start = 0;
    y_start = 0;
    per_pixel = +1;
    per_src_line = m_area.w;
    per_dest_line = m_area.w;
  }
  else if (source.x >= dest.x && source.y <= dest.y) {
    x_start = dest.w - 1;
    y_start = 0;
    per_pixel = -1;
    per_src_line = m_area.w;
    per_dest_line = m_area.w;
  }
  else if (source.x <= dest.x && source.y >= dest.y) {
    x_start = 0;
    y_start = dest.h - 1;
    per_pixel = +1;
    per_src_line = -m_area.w;
    per_dest_line = -m_area.w;
  }
  else {
    x_start = dest.w - 1;
    y_start = dest.h - 1;
    per_pixel = -1;
    per_src_line = -m_area.w;
    per_dest_line = -m_area.w;
  }
  
  coord dest_pixel = ((dest.y + y_start) * m_area.w) + dest.x + x_start;
  coord src_pixel = ((source.y + y_start) * m_area.w) + source.x + x_start;
  
  InternalBitBlt (dest.w, dest.h, per_pixel, per_src_line, per_dest_line,
		  &m_data [src_pixel], &m_data [dest_pixel] );
  
}

void GsmpPixmap::BitBlt (const GsmpPixmap& pixmap, GsmpPixmap::area source,
			 const GsmpPixmap::point& dest_point)
{
  // self bitblt-ing?
  if (&pixmap == this) {
    cout << "GsmpPixmap::BitBlt: use of wrong method! - dynamically corrected ..." << endl;
    BitBlt (source, dest_point);
  }
  else {
    source.Clip (pixmap.Area () ); // clip to source boundaries
    area dest (dest_point.x, dest_point.y, source.w, source.h);
    dest.Clip (m_area); // clip destination boundaries
    
    cout << "GsmpPixmap::BitBlt: source: " << source << ", dest: " << dest << endl;
    
    coord dest_pixel = (dest.y * m_area.w) + dest.x;
    coord src_pixel = (source.y * pixmap.m_area.w) + source.x;
    
    InternalBitBlt (dest.w, dest.h, +1, pixmap.m_area.w, m_area.w,
		    &pixmap.m_data [src_pixel], &m_data [dest_pixel] );
  }
}

void GsmpPixmap::DrawLine (GsmpPixmap::point p1, GsmpPixmap::point p2,
			   GsmpColor::Color c)
{
  if (! InternalClipLine (p1, p2) )
    return;
  
  coord dx = p2.x - p1.x;
  coord dy = p2.y - p1.y;
  cout << "DrawLine: " << p1 << " -> " << p2 << ", dx: " << dx << ", dy: " << dy << endl;
  
  if (dx == 0) // vertical line
    {
      coord y = std::min (p1.y, p2.y);
      coord t_y = std::max (p1.y, p2.y);
      
      coord pixel = (y * m_area.w) + p1.x;
      
      while (y <= t_y){
	GsmpColor::Blend ( &(m_data [pixel] ), c);
	pixel += m_area.w;
	++ y;
      }
    }
  else if (dy == 0) // horizontal line
    {
      coord x = std::min (p1.x, p2.x);
      coord t_x = std::max (p1.x, p2.x);
      
      coord pixel = (p1.y * m_area.w) + x;
      
      while (x <= t_x){
	GsmpColor::Blend ( &(m_data [pixel] ), c);
	++ pixel;
	++ x;
      }
    }
  // here we draw a line using the midpoint scan-conversion algorithm
  else
    {
      if (p2.y > p1.y) { // positive angle
	if (dy <= dx) // flat
	  {
	    coord d = 2 * dy - dx;
	    coord incE = 2 * dy;
	    coord incNE  = 2 * (dy - dx);
	    
	    coord pixel = (p1.y * m_area.w) + p1.x;
	    GsmpColor::Blend ( &(m_data [pixel] ), c);
	    
	    while (p1.x < p2.x)
	      {
		if (d <= 0) {
		  d += incE;
		  ++ pixel;
		}
		else {
		  d += incNE;
		  pixel += m_area.w + 1;
		}
		GsmpColor::Blend ( &(m_data [pixel] ), c);
		++ p1.x;
	      }
	  }
	else // steep (this is a late night guess and might need fixing)
	  {
	    coord d = 2 * dx - dy;
	    coord incE = 2 * dx;
	    coord incNE  = 2 * (dx - dy);
	    
	    coord pixel = (p1.y * m_area.w) + p1.x;
	    GsmpColor::Blend ( &(m_data [pixel] ), c);
	    
	    while (p1.y < p2.y)
	      {
		if (d <= 0) {
		  d += incE;
		  pixel += m_area.w;
		}
		else {
		  d += incNE;
		  pixel += m_area.w + 1;
		}
		GsmpColor::Blend ( &(m_data [pixel] ), c);
		++ p1.y;
	      }
	  }
      }
      else { // negative angle (this is also late night guess and might need fixing)
	if (-dy <= dx) // flat
	  {
	    coord d = 2 * (-dy) - dx;
	    coord incE = 2 * (-dy);
	    coord incNE  = -2 * (dy + dx);
	    
	    coord pixel = (p1.y * m_area.w) + p1.x;
	    GsmpColor::Blend ( &(m_data [pixel] ), c);
	    
	    while (p1.x < p2.x)
	      {
		if (d <= 0) {
		  d += incE;
		  ++ pixel;
		}
		else {
		  d += incNE;
		  pixel -= m_area.w - 1;
		}
		GsmpColor::Blend ( &(m_data [pixel] ), c);
		++ p1.x;
	      }
	  }
	else // steep (the last late night guess and might need fixing)
	  {
	    coord d = 2 * (-dx) - dy;
	    coord incE = 2 * dx;
	    coord incNE  = 2 * (dx + dy);
	    
	    coord pixel = (p1.y * m_area.w) + p1.x;
	    GsmpColor::Blend ( &(m_data [pixel] ), c);
	    
	    while (p1.y > p2.y)
	      {
		if (d <= 0) {
		  d += incE;
		  pixel -= m_area.w;
		}
		else {
		  d += incNE;
		  pixel -= m_area.w - 1;
		}
		GsmpColor::Blend ( &(m_data [pixel] ), c);
		-- p1.y;
	      }
	  }
      }
    }
}

void GsmpPixmap::DrawAaLine (GsmpPixmap::point p1, GsmpPixmap::point p2,
			     GsmpColor::Color c)
{
  if (! InternalClipLine (p1, p2) )
    return;
  
  coord dx = p2.x - p1.x;
  coord dy = p2.y - p1.y;
  cout << "DrawLine: " << p1 << " -> " << p2 << ", dx: " << dx << ", dy: " << dy << endl;
  
  if (dx == 0) // vertical line
    {
      coord y = std::min (p1.y, p2.y);
      coord t_y = std::max (p1.y, p2.y);
      
      coord pixel = (y * m_area.w) + p1.x;
      
      while (y <= t_y){
	GsmpColor::Blend ( &(m_data [pixel] ), c);
	pixel += m_area.w;
	++ y;
      }
    }
  else if (dy == 0) // horizontal line
    {
      coord x = std::min (p1.x, p2.x);
      coord t_x = std::max (p1.x, p2.x);
      
      coord pixel = (p1.y * m_area.w) + x;
      
      while (x <= t_x){
	GsmpColor::Blend ( &(m_data [pixel] ), c);
	++ pixel;
	++ x;
      }
    }
  // here we draw a line using the midpoint scan-conversion algorithm
  else
    {
      if (p2.y > p1.y) { // positive angle
	if (dy <= dx) // flat
	  {
	    coord d = 2 * dy - dx;
	    coord incE = 2 * dy;
	    coord incNE  = 2 * (dy - dx);
	    
	    coord pixel = (p1.y * m_area.w) + p1.x;
	    
	    while (p1.x < p2.x)
	      {
		coord intB = ((d - incNE) << 8) / (incE - incNE + 1);
		coord intA = 255 - intB;
		
		GsmpColor::Blend ( &(m_data [pixel] ), c, intA);
		if (pixel + m_area.w < m_area.w * m_area.h)
		  GsmpColor::Blend ( &(m_data [pixel + m_area.w] ), c, intB);
		
		if (d <= 0) {
		  d += incE;
		  ++ pixel;
		}
		else {
		  d += incNE;
		  pixel += m_area.w + 1;
		}
		
		++ p1.x;
	      }
	  }
	else // steep (this is a late night guess and might need fixing)
	  {
	    coord d = 2 * dx - dy;
	    coord incE = 2 * dx;
	    coord incNE  = 2 * (dx - dy);
	    
	    coord pixel = (p1.y * m_area.w) + p1.x;
	    
	    while (p1.y < p2.y)
	      {
		coord intB = ((d - incNE) << 8) / (incE - incNE + 1);
		coord intA = 255 - intB;
		
		GsmpColor::Blend ( &(m_data [pixel] ), c, intA);
		if (pixel + 1 < m_area.w * m_area.h) // TODO: needs fixing
		  GsmpColor::Blend ( &(m_data [pixel + 1] ), c, intB);
		
		if (d <= 0) {
		  d += incE;
		  pixel += m_area.w;
		}
		else {
		  d += incNE;
		  pixel += m_area.w + 1;
		}
		
		++ p1.y;
	      }
	  }
      }
      else { // negative angle (this is also late night guess and might need fixing)
	if (-dy <= dx) // flat
	  {
	    coord d = 2 * (-dy) - dx;
	    coord incE = 2 * (-dy);
	    coord incNE  = -2 * (dy + dx);
	    
	    coord pixel = (p1.y * m_area.w) + p1.x;
	    
	    while (p1.x < p2.x)
	      {
		coord intB = ((d - incNE) << 8) / (incE - incNE + 1);
		coord intA = 255 - intB;
		
		GsmpColor::Blend ( &(m_data [pixel] ), c, intA);
		if (pixel - m_area.w >= 0)
		  GsmpColor::Blend ( &(m_data [pixel - m_area.w] ), c, intB);
		
		if (d <= 0) {
		  d += incE;
		  ++ pixel;
		}
		else {
		  d += incNE;
		  pixel -= m_area.w - 1;
		}
		
		++ p1.x;
	      }
	  }
	else // steep (the last late night guess and might need fixing)
	  {
	    coord d = 2 * (-dx) - dy;
	    coord incE = 2 * dx;
	    coord incNE  = 2 * (dx + dy);
	    
	    coord pixel = (p1.y * m_area.w) + p1.x;
	    
	    while (p1.y > p2.y)
	      {
		coord intB = ((d - incNE) << 8) / (incE - incNE + 1);
		coord intA = 255 - intB;
		
		GsmpColor::Blend ( &(m_data [pixel] ), c, intA);
		if (pixel + 1 < m_area.w * m_area.h) // TODO: needs fixing
		  GsmpColor::Blend ( &(m_data [pixel + 1] ), c, intB);
		
		if (d <= 0) {
		  d += incE;
		  pixel -= m_area.w;
		}
		else {
		  d += incNE;
		  pixel -= m_area.w - 1;
		}
		
		-- p1.y;
	      }
	  }
      }
    }
}

void GsmpPixmap::DrawRectangle (const GsmpPixmap::area& a, GsmpColor::Color c)
{
  cout << "Rectangle: " << a << endl;
  DrawLine (point (a.x, a.y), point (a.x, a.y + a.h - 1), c);
  DrawLine (point (a.x, a.y), point (a.x + a.w - 1, a.y), c);
  DrawLine (point (a.x + a.w - 1, a.y), point (a.x + a.w - 1, a.y + a.h - 1), c);
  DrawLine (point (a.x, a.y + a.h - 1), point (a.x + a.w - 1, a.y + a.h - 1), c);
}

void GsmpPixmap::DrawFilledRectangle (const GsmpPixmap::area& a, GsmpColor::Color c)
{
  cout << "FilledRectangle: " << a << endl;
  area b = a;
  b.Clip (m_area);
  
  cout << "FilledRectangle: clipped: " << b << endl;
  
  for (int i = b.y; i < b.y + b.h; ++ i)
    DrawLine (point (b.x, i), point (b.x + b.w - 1, i), c);
}

void GsmpPixmap::DrawCircle (const GsmpPixmap::point& p1, GsmpPixmap::coord r,
			     GsmpColor::Color c)
{
  cout << "GsmpPixmap::DrawCircle: not implemented ..." << endl;
}

void GsmpPixmap::DrawFilledCircle (const GsmpPixmap::point& p1, GsmpPixmap::coord r,
				   GsmpColor::Color c)
{
  cout << "GsmpPixmap::DrawFilledCircle: not implemented ..." << endl;

}

void GsmpPixmap::DrawCircle (const GsmpPixmap::point& p1, GsmpPixmap::coord r,
			     GsmpPixmap::coord angle1, GsmpPixmap::coord angle2,
			     GsmpColor::Color c)
{
  cout << "GsmpPixmap::DrawCircle: not implemented ..." << endl;
}

void GsmpPixmap::DrawFilledCircle (const GsmpPixmap::point& p1, GsmpPixmap::coord r,
				   GsmpPixmap::coord angle1, GsmpPixmap::coord angle2,
				   GsmpColor::Color c)
{
  cout << "GsmpPixmap::DrawFilledCircle: not implemented ..." << endl;
}

void GsmpPixmap::DrawAaCircle (const GsmpPixmap::point& p1, GsmpPixmap::coord r,
			       GsmpColor::Color c)
{
  cout << "GsmpPixmap::DrawAaCircle: not implemented ..." << endl;
}

void GsmpPixmap::DrawAaCircle (const GsmpPixmap::point& p1, GsmpPixmap::coord r,
			       GsmpPixmap::coord angle1, GsmpPixmap::coord angle2,
			       GsmpColor::Color c)
{
  cout << "GsmpPixmap::DrawAaCircle: not implemented ..." << endl;
}

void GsmpPixmap::DrawEllipse (const GsmpPixmap::point& p, GsmpPixmap::coord a,
			      GsmpPixmap::coord b, GsmpColor::Color c)
{
  cout << "GsmpPixmap::DrawEllipse: not implemented ..." << endl;
}

void GsmpPixmap::DrawAaEllipse (const GsmpPixmap::point& p, GsmpPixmap::coord a,
				GsmpPixmap::coord b, GsmpColor::Color c)
{
  cout << "GsmpPixmap::DrawAaEllipse: not implemented ..." << endl;
}

inline bool GsmpPixmap::InternalClipLine (point& p1, point& p2)
{
  // trival reject
  if ( ((p1.y < 0) && (p2.y < 0)) ||
       ((p1.y >= m_area.h) && (p2.y >= m_area.h)) ||
       ((p1.x < 0) && (p2.x < 0)) ||
       ((p1.x >= m_area.w) && (p2.x >= m_area.w)) )
    {
      cout << "InternalClipLine: trival reject" << endl;
      return false;
    }
  // trival accept or clipping?
  if (!m_area.Contains (p1) || !m_area.Contains (p2) )
    {
      cout << "InternalClipLine: performing clipping" << endl;
      /* clip to top edge */
      if (p1.y < 0)
	{
	  p1.x += (p1.y * (p1.x - p2.x)) / (p2.y - p1.y);
	  p1.y = 0;
	}
      if (p2.y < 0)
	{
	  p2.x += (p2.y * (p1.x - p2.x)) / (p2.y - p1.y);
	  p2.y = 0;
	}
    
      /* clip to bottom edge */
      if (p1.y >= m_area.h)
	{
	  p1.x -= ((m_area.h - p1.y) * (p1.x - p2.x)) / (p2.y - p1.y);
	  p1.y = m_area.h - 1;
	}
      if (p2.y >= m_area.h)
	{
	  p2.x -= ((m_area.h - p2.y) * (p1.x - p2.x)) / (p2.y - p1.y);
	  p2.y = m_area.h - 1;
	}
    
      /* clip to left edge */
      if (p1.x < 0)
	{
	  p1.y += (p1.x * (p1.y - p2.y)) / (p2.x - p1.x);
	  p1.x = 0;
	}
      if (p2.x < 0)
	{
	  p2.y += (p2.x * (p1.y - p2.y)) / (p2.x - p1.x);
	  p2.x = 0;
	}
    
      /* clip to right edge */
      if (p1.x >= m_area.w)
	{
	  p1.y -= ((m_area.w - p1.x) * (p1.y - p2.y)) / (p2.x - p1.x);
	  p1.x = m_area.w - 1;
	}
      if (p2.x >= m_area.w)
	{
	  p2.y -= ((m_area.w - p2.x) * (p1.y - p2.y)) / (p2.x - p1.x);
	  p2.x = m_area.w - 1;
	}
    }
  
  /* make sure p1 is the leftmost */
  if (p1.x > p2.x) {
    point t = p2;
    p2 = p1;
    p1 = t;
  }
  
  return true;
}

inline void GsmpPixmap::InternalBitBlt (coord x_count, coord y_count,
					coord per_pixel, coord per_src_line, coord per_dest_line,
					GsmpColor::Color* src, GsmpColor::Color* dest)
{
  for (coord y = 0; y < y_count; ++ y)
    {
      GsmpColor::Color* t_dest = dest;
      GsmpColor::Color* t_src = src;
      
      for (coord x = 0; x < x_count; ++ x) {
	GsmpColor::Blend (t_dest, *t_src);
	t_dest += per_pixel;
	t_src += per_pixel;
      }
      
      dest += per_dest_line;
      src += per_src_line;
    }
}

/************
 * GsmpFace *
 ************/

GsmpFace::GsmpFace (GsmpFontSystem* i_system, FT_Library* i_library)
{
  m_system = i_system;
  m_library = i_library;
  m_loaded = false;
}

GsmpFace::~GsmpFace ()
{
  if (m_loaded)
    FT_Done_Face (m_face);
}

bool GsmpFace::LoadFace (const string& family, const string& style)
{
  if (m_loaded)
    return false;
  
  m_filename = m_system->GetFileForFace (family, style);
  
  if (m_filename == "")
    return false;
  
  int error = FT_New_Face (*m_library,
			   m_filename.c_str (),
			   0,
			   &m_face);
  
  if (error == FT_Err_Unknown_File_Format) {
    cout << "font format not supported ..." << endl;
    return false;
    
  }
  else if (error) {
    cout << "font file doesn't exist or is broken ..." << endl;
    return false;
  }
  
  cout << "family_name: " << m_face->family_name
       << ", style name: " << m_face->style_name << endl;
  
  m_loaded = true;
}

bool GsmpFace::UnloadFace ()
{
  if (!m_loaded)
    return false;
  
  FT_Done_Face (m_face);
  m_loaded = false;
  return true;
}

void GsmpFace::Test (GsmpPixmap* t_pixmap)
{
  if (!m_loaded)
    return;
  
  int error;
  error = FT_Set_Char_Size(m_face,  /* handle to face object           */
			   0,       /* char_width in 1/64th of points  */
			   11*64,   /* char_height in 1/64th of points */
			   72,      /* horizontal device resolution    */
			   72);     /* vertical device resolution      */
  if (error) {
    cout << "error setting the charackter size ..." << endl;
    return;
  }
  
  string text = "GsmpText! AaBb ;-}";
  
  int pen_x, pen_y;
  pen_x = pen_y = 20;
  
  for (int i = 0; i < text.size (); ++ i)
    {
      cout << "rendering: " << text [i] << endl;
      FT_UInt  glyph_index;
         
      // retrieve glyph index from character code
      glyph_index = FT_Get_Char_Index (m_face, text[i]);
      
      // load glyph image into the slot (erase previous one)
      error = FT_Load_Glyph (m_face, glyph_index, FT_LOAD_DEFAULT);
      if (error) {
	cout << "error loading glyph ..." << endl;
	continue;  // ignore errors
      }
      
      // convert to an anti-aliased bitmap
      error = FT_Render_Glyph (m_face->glyph, ft_render_mode_normal);
      if (error) {
	cout << "error rendering glyph ..." << endl;
	continue;
      }
      
      FT_GlyphSlot slot = m_face->glyph;  // a small shortcut
      
      cout << "bitmap_left: " << slot->bitmap_left
	   << ", bitmap_top: " << slot->bitmap_top << endl;
      
      cout << "width: " << slot->bitmap.width << ", rows: " << slot->bitmap.rows
	   << ", pitch: " << slot->bitmap.pitch
	   << ", num_grays: " << slot->bitmap.num_grays << endl;
      
      DrawTfBitmap (t_pixmap, &slot->bitmap,
		    pen_x + slot->bitmap_left,
		    pen_y - slot->bitmap_top);
      
      // increment pen position 
      pen_x += slot->advance.x >> 6;
      pen_y += slot->advance.y >> 6;   // unuseful for now..
    }
}

void GsmpFace::DrawTfBitmap (GsmpPixmap* t_dest, FT_Bitmap_* t_src,
			     GsmpPixmap::coord dest_x, GsmpPixmap::coord dest_y)
{
  GsmpColor::Color* pixel = t_dest->m_data + (t_dest->m_area.w * dest_y) + dest_x;
  
  cout << "dest_x: " << dest_x << ", dest_y: " << dest_y << endl;
  
  for (int y = 0; y < t_src->rows; ++y) {
    for (int x = 0; x < t_src->width; ++x) {
      GsmpColor::Component alpha = *(t_src->buffer + (t_src->width * y) + x);
      if (alpha != 0) /* saves some cpu cycles */
	GsmpColor::Blend (pixel + x, 0xffff0000, alpha);
    }
    pixel += t_dest->m_area.w;
  }
}

/******************
 * GsmpFontSystem *
 ******************/

GsmpFontSystem::GsmpFontSystem ()
{
  m_initilized = false;
}

GsmpFontSystem::~GsmpFontSystem ()
{
  // TODO: delete FontEntry structs and free FreeType structs
}

bool GsmpFontSystem::Initialize ()
{
  m_initilized = ! FT_Init_FreeType (&m_library);
  
  if (!m_initilized) {
    cout << "GsmpFont::Initialize: Error during library initialization!" << cout;
    return false;
  }
  
  // try to load all files and create index
  ScanFontDir ("/usr/X11/lib/X11/fonts");
  ScanFontDir ("/usr/share/ghostscript/fonts");
  ScanFontDir ("/usr/share/fonts/default/ghostscript");
  return m_initilized;
}

GsmpFontSystem::const_iterator GsmpFontSystem::Begin () const
{
  return m_fonts.begin ();
}

GsmpFontSystem::const_iterator GsmpFontSystem::End () const
{
  return m_fonts.end ();
}

GsmpFace* GsmpFontSystem::NewFace ()
{
  GsmpFace* t_face = new GsmpFace (this, &m_library);
  return t_face;
}

string GsmpFontSystem::GetFileForFace (const string& family,
				       const string& style) const
{
  string t_file;
  bool t_found = false;
  
  for (const_iterator i = m_fonts.begin (); !t_found && i != m_fonts.end(); ++i)
    {
      if (i->family == family && i->style == style) {
	cout << "found font" << endl;
	t_found = true;
	t_file = i->filename;
      }
    }
  
  cout << "returning: " << t_file << endl;
  return t_file;
}

// ***

bool GsmpFontSystem::ScanFontDir (const string& dir)
{
  if (!m_initilized)
    return false;
  
  DirIterator t_dir;
  
  if (!t_dir.Open (dir) )
    return false;
  
  Timer t_timer;
  int t_scanned_files = 0;
  
  for (; ! t_dir.End (); ++ t_dir)
    {
      string file = *t_dir;
      string ext;
      
      ++ t_scanned_files;
      
      string::size_type idx = file.rfind ('.');
      if (idx != string::npos)
	ext = file.substr (idx + 1);
      
      cout << "File: " << file << ", ext: " << ext << endl;
      
      if (file == "." || file == "..")
	continue;
      
      file = dir + "/" + file;
      
      DirIterator t_test_dir;
      if (t_test_dir.Open (file) ) // new directory? -> recursion
	{
	  cout << "Directory -> scan: " << file << endl;
	  t_test_dir.Close ();
	  ScanFontDir (file);
	}
      else // open file
	{
	  /* Supported by FreeType2:
	   * TrueType fonts (and collections)
	   * Type 1 fonts
	   * CID-keyed Type 1 fonts
	   * CFF fonts
	   * OpenType fonts (both TrueType and CFF variants)
	   * SFNT-based bitmap fonts
	   * X11 PCF fonts
	   * Windows FNT fonts
	   */
	  
	  if (ext == "pfa" || ext == "ttf" || ext == "TTF")
	    {
	      FT_Face t_face; 
	      int error = FT_New_Face (m_library,
				       file.c_str (),
				       0,
				       &t_face);
	      
	      if (error == FT_Err_Unknown_File_Format) {
		cout << "font format not supported ..." << endl;
	      }
	      else if (error) {
		cout << "font file doesn't exist or is broken ..." << endl;
	      }
	      
	      FontEntry t_font;
	      t_font.family = t_face->family_name;
	      t_font.style = t_face->style_name;
	      t_font.filename = file;
	      m_fonts.push_back (t_font);
	      
	      FT_Done_Face (t_face);
	    }
	  
	} // end open file
    } // end while next file
  
  cout << t_scanned_files << " file,  scanned in "
       << t_timer.Delta () << " seconds." << endl;
}

void GsmpFontSystem::Debug ()
{
  cout << "GsmpFontSystem::Debug" << endl;
  if (!m_initilized) {
    cout << "  not initialized" << endl;
  }
  else {
    int count = 0;
    for (iterator i = m_fonts.begin (); i != m_fonts.end(); ++i, ++count)
      {
	cout << "\"" << i->family << "\" \"" << i->style
	     << "\" \"" << i->filename << "\"" << endl;
      }
    cout << count << " Fonts." << endl;
  }
}
